﻿Imports System.ComponentModel
Imports System.Data.Common
Imports System.Text.RegularExpressions
Imports System.Runtime.CompilerServices

Public MustInherit Class DbPersistanceBase : Inherits WeakAttachBase(Of DbPersistanceBase)

	Public ReadOnly Property Connection As DbConnection
	Public RankedTables As New List(Of DataTable)
	Public ReadOnly DatasetClone As DataSet
	Public ReadOnly Quotes As String()

	Public ReadOnly DataSet As DataSet
	Private _ForeignKeyStatui As List(Of Tuple(Of ForeignKeyConstraint, AcceptRejectRule))

	Protected MustOverride Sub InitAutoIncrement(table As DataTable, adp As DbDataAdapter)

	Public Sub New(con As DbConnection, dts As DataSet, factory As DbProviderFactory)
		If NotifyPropertyChanged.IsInDesignMode Then Return
		DbPersistanceBase.AddAttach(dts, Me)
		Connection = con
		DataSet = dts
		DatasetClone = dts.Clone
		DatasetClone.EnforceConstraints = False
		'Store the AcceptRejectRule of all datasets ForeignKeyConstraints
		_ForeignKeyStatui = Aggregate rl In DataSet.Relations.Cast(Of DataRelation)
								Let fk = rl.ChildKeyConstraint Where fk IsNot Nothing
								Select Tuple.Create(fk, fk.AcceptRejectRule) Into ToList
		BuildRankedTables()
		Connection.OpenUntilIdle
		Dim quoteInfo = Connection.GetSchema("DataSourceInformation").Rows(0)("QuotedIdentifierPattern").ToString.Replace("\", "")
		Dim i = quoteInfo.IndexOf("|"c)
		Quotes = {quoteInfo(i - 2), quoteInfo(i + 1)}
		For Each tb In RankedTables
			Dim cmd = Connection.CreateCommand
			Dim tabName = tb.TableName.Between(Quotes)
			Dbg(tabName)
			cmd.CommandText = $"SELECT {tabName}.* FROM {tabName}"
			Dim adp = factory.CreateDataAdapter
			adp.SelectCommand = cmd
			Using builder = factory.CreateCommandBuilder
				With builder
					.DataAdapter = adp
					.QuotePrefix = Quotes(0) : .QuoteSuffix = Quotes(1)
					.ConflictOption = ConflictOption.OverwriteChanges
				End With
				With adp
					.UpdateCommand = DirectCast(DirectCast(builder.GetUpdateCommand, ICloneable).Clone, DbCommand)
					.InsertCommand = DirectCast(DirectCast(builder.GetInsertCommand, ICloneable).Clone, DbCommand)
					.DeleteCommand = DirectCast(DirectCast(builder.GetDeleteCommand, ICloneable).Clone, DbCommand)
					.TableMappings.Add(DbDataAdapter.DefaultSourceTableName, tb.TableName)
					.MissingSchemaAction = MissingSchemaAction.Ignore
				End With
			End Using
			If tb.PrimaryKey.Length = 1 AndAlso tb.PrimaryKey(0).AutoIncrement Then InitAutoIncrement(tb, adp)
			Dim tp = New TablePersist(tb, adp)
		Next
	End Sub

	Public Sub New(con As DbConnection, dts As DataSet)
		Me.New(con, dts, DbProviderFactories.GetFactory(con.GetType.Namespace))
	End Sub

	''' <summary>collect Tables into _RankedTables in topologic order: no childtable is before its parentTable</summary>
	Private Sub BuildRankedTables()
		Dim toDo = New HashSet(Of DataTable)(DataSet.TablesX.Where(Function(tb) tb.PrimaryKey.Length > 0))
		Dim recurse As Action(Of DataTable) =
			Sub(tb)
				' before each table put into _RankedTables put all its parentTables
				If Not toDo.Remove(tb) Then Return 'prevent run in Cycles
				For Each rl As DataRelation In tb.ParentRelations
					If rl.ChildKeyConstraint IsNot Nothing Then recurse(rl.ParentTable)
				Next
				RankedTables.Add(tb)
			End Sub
		While toDo.Count > 0 : recurse(toDo.First) : End While ' repeat call recurse(), until all tables are done
	End Sub

	Public Sub Save()
		' process all tables twice, each time in ranked order. First select and update Addeds and Modifieds
		' After that update the whole tables (where only the Deleteds remain change-tracked) - while temporally all ForeignKeys are set to .AcceptRejectRule.Cascade
		DataSet.AcceptVirtualChanges()
		Dim s = DataSet.ReportOverview(exclude:=DataRowState.Unchanged)
		If s <> "" Then Dbg($"Save {s}")
		Connection.OpenUntilIdle
		Dim count = 0
		Const AddedsAndModifieds = DataViewRowState.Added Or DataViewRowState.ModifiedCurrent
		RankedTables.ForEach(Sub(tb) count += tb.TablePersist.Update(tb.Select("", "", AddedsAndModifieds)))
		' set all FKs.AcceptRejectRule.Cascade. This causes DataAdapters to skip rows, which were deleted by delete-cascade
		_ForeignKeyStatui.ForEach(Sub(fk) fk.Item1.AcceptRejectRule = AcceptRejectRule.Cascade)
		RankedTables.ForEach(Sub(tb) count += tb.TablePersist.Update()) ' update Deleteds 
		_ForeignKeyStatui.ForEach(Sub(fk) fk.Item1.AcceptRejectRule = fk.Item2) ' restore FKs.AcceptRejectRule
	End Sub

End Class
