﻿Imports System.Runtime.CompilerServices
Imports System.ComponentModel, System.Diagnostics
Imports System.IO

Namespace Global.System.Data
	<EditorBrowsable(EditorBrowsableState.Never)>
	Public Module DataX

#Region "Dataset"

		''' <summary> reset ChangeTracking of modified Datarows, which are not to send to the Database.
		''' This affects changes at ForeignKeyColumns as well as Changes made undone by the user.</summary>
		<Extension()>
		Public Sub AcceptVirtualChanges(dts As DataSet)
			Dim changes = 0, virtualChanges = 0 ' counter for debug
			Dim getCurrentValues = Function(cols As DataColumn(), rw As DataRow) cols.Select(Function(col) rw(col, DataRowVersion.Current))
			Dim getOriginalValues = Function(cols As DataColumn(), rw As DataRow) cols.Select(Function(col) rw(col, DataRowVersion.Original))
			For Each tb As DataTable In dts.Tables
				Dim modifiedRows = tb.Select("", "", DataViewRowState.ModifiedCurrent)
				If modifiedRows.Length = 0 Then Continue For
				changes += modifiedRows.Length
				Dim fkRels = Aggregate rl In tb.ParentRelations.Cast(Of DataRelation) Where rl.ChildKeyConstraint IsNot Nothing Into ToList
				Dim fkColumns = fkRels.SelectMany(Function(rl) rl.ChildColumns).ToArray
				Dim payloadColumns = tb.Columns.Cast(Of DataColumn).Where(Function(col) col.Expression = "").Except(fkColumns).ToArray
				For Each rw In modifiedRows
					'rw-ChangeTracking rücksetzen, wenn in allen Nicht-FK-Spalten ('payload') sich aktueller und Original-Wert gleichen
					'und ausserdem für alle FK-Relationen sich Original- und Parent-Original-Wert gleichen.
					'letzteres markiert einen 'VirtualChange': die Spalte wurde nicht explizit geändert, sondern von der ParentRow
					' wurde die Änderung aufgrund DataRelation.UpdateRule.Cascade injiziert
					If getCurrentValues(payloadColumns, rw).SequenceEqual(getOriginalValues(payloadColumns, rw)) Then
						Dim allOrigParentValues = fkRels.SelectMany(Function(rl) getOriginalValues(rl.ParentColumns, rw.GetParentRow(rl)))
						If getOriginalValues(fkColumns, rw).SequenceEqual(allOrigParentValues) Then rw.AcceptChanges() : virtualChanges += 1
					End If
				Next
			Next
			'Debug.WriteLine($"AcceptVirtualChanges(): changes={changes}, remainingChanges={changes - virtualChanges}")
		End Sub

		''' <summary> for debugging-purposes </summary>
		<Extension()>
		Public Function ReportOverview(dts As DataSet, Optional exclude As DataRowState = Nothing) As String
			Dim entries = New List(Of String)
			For Each tb In dts.TablesX
				Dim rwGroups = tb.RowsX.ToLookup(Function(x) x.RowState)
				Dim grps = Aggregate rw In tb.RowsX Where Not exclude.HasFlag(rw.RowState) Group By rw.RowState Into Count Into ToList
				If grps.Count > 0 Then
					entries.Add($"{tb.TableName}: " & " ".Between(grps.Select(Function(x) $"{x.RowState}_{x.Count}")))
				End If
			Next
			Return ", ".Between(entries)
		End Function
		''' <summary>legt für alle Foreignkey-Spalten im Dataset den DefaultValue auf eine (beliebige) existierende ParentRow fest.
		''' Befördert die Validität via DataTable.NewRow() erzeugter Datarows</summary>
		<Extension()>
		Public Sub EnsureRelationDefaults(dts As DataSet)
			For Each tbl In dts.GetRankedTables
				Dim rwDefault = tbl.Rows.Cast(Of DataRow).FirstOrDefault
				If rwDefault.Null Then Continue For
				For Each rl As DataRelation In tbl.ChildRelations
					rl.ParentColumns.Zip(rl.ChildColumns).ForEach(Sub(tpl) tpl.Item2.DefaultValue = rwDefault(tpl.Item1))
				Next
			Next
		End Sub
		''' <summary>topologic sort: no childtable is before its parentTable</summary>
		<Extension()>
		Public Function GetRankedTables(dts As DataSet) As List(Of DataTable)
			Dim hsh = New HashSet(Of DataTable)(dts.Tables.Cast(Of DataTable))
			GetRankedTables = New List(Of DataTable)
			Dim recurse As Action(Of DataTable) =
				Sub(tb)
					If Not hsh.Remove(tb) Then Return
					tb.ParentTables.ForEach(recurse)
					GetRankedTables.Add(tb)
				End Sub
			While hsh.Count > 0 : recurse(hsh.First()) : End While
		End Function

		''' <summary>for Debug/Testing: converts each record to a string</summary>
		<Extension()>
		Public Iterator Function DumpDataset(dts As DataSet) As IEnumerable(Of String)
			For Each tb As DataTable In dts.Tables
				For Each rw As DataRow In tb.All
					Yield "·".Between(rw.ItemArray)
				Next
			Next
		End Function

#Region "ReadWriteCompact"
		'Das Dataset wird durchlaufen, und alle Werte per TypConverter von und zu String umgewandelt. Schema-Validierung besteht darin, dass die Dataset-Struktur genau definiert, wann welcher TypConverter verwendet wird. Die Datei ist nur 1/7 so gross, v.a., weil sie keine Schema-Info enthält. Entscheidend aber, dass viel mehr Datentypen persistierbar sind als per Dataset.WriteXml - zB Fonts, Points und alle Enums - halt alles, wofür TypeConverter implementiert sind.


		Private Class DataRowConverter
			Private _KVPs As New List(Of Tuple(Of DataColumn, AllConverter)), Table As DataTable

			Public Sub New(Table As DataTable)
				Me.Table = Table
				_KVPs = (
					From c In Table.Columns.Cast(Of DataColumn)() Where c.Expression = ""
					Select Tuple.Create(c, New AllConverter(c.DataType))).ToList
			End Sub
			Public Function Parse(data As String()) As DataRow
				If data.Length <> _KVPs.Count Then Throw Me.Exception("unexpected count of input-strings")
				Dim rw = Table.NewRow
				For i = 0 To _KVPs.Count - 1
					Dim s = data(i)
					With _KVPs(i)
						rw(.Item1) = If(s = CompactReadWrite.DBNull, Convert.DBNull, .Item2.FromString(s))
					End With
				Next
				Return rw
			End Function
			Public Function Format(Row As DataRow) As String()
				Dim strings(_KVPs.Count - 1) As String
				For i = 0 To _KVPs.Count - 1
					With _KVPs(i)
						Dim val = Row(.Item1)
						strings(i) = If(Convert.IsDBNull(val), CompactReadWrite.DBNull, .Item2.ToString(val))
					End With
				Next
				Return strings
			End Function
		End Class

		''' <summary>reads from file, expecting the compact-file-format</summary>
		<Extension()>
		Public Sub LoadCompact(Subj As DataSet, dataFile As String)
			Subj.LoadCompact(New StreamReader(dataFile))
		End Sub
		''' <summary>reads from TextReader, expecting the compact-file-format</summary>
		<Extension()>
		Public Sub LoadCompact(Subj As DataSet, reader As TextReader, Optional LeaveWriterOpen As Boolean = False)
			Subj.Clear()
			Dim rankedTables = Subj.GetRankedTables.ToList
			rankedTables.ForEach(Sub(tb) tb.BeginLoadData())
			Using r = New CompactReadWrite(reader, LeaveWriterOpen)
				For Each tb In rankedTables
					Dim converter = New DataRowConverter(tb)
					r.ReadTable.Select(AddressOf converter.Parse).ForEach(AddressOf tb.Rows.Add)
				Next
			End Using
			'rankedTables.ForEach(Sub(tb) tb.EndLoadData())
			For i = rankedTables.Count - 1 To 0 Step -1
				rankedTables(i).EndLoadData()
			Next
		End Sub

		''' <summary>writes to file in compact-file-format</summary>
		<Extension()>
		Public Sub SaveCompact(Subj As DataSet, dataFile As String)
			Subj.SaveCompact(New StreamWriter(dataFile))
		End Sub
		''' <summary>writes to TextWriter in compact-file-format</summary>
		<Extension()>
		Public Sub SaveCompact(Subj As DataSet, writer As TextWriter, Optional LeaveWriterOpen As Boolean = False)
			Using w = New CompactReadWrite(writer, LeaveWriterOpen)
				For Each tb In Subj.GetRankedTables
					w.WriteHeader(tb.TableName,
						From c In tb.Columns.Cast(Of DataColumn)() Where c.Expression = "" Select c.ColumnName)
					Dim converter = New DataRowConverter(tb)
					tb.All.Select(AddressOf converter.Format).ForEach(AddressOf w.WriteRow)
					w.CloseWriteTable()
				Next
			End Using
		End Sub

#End Region 'ReadWriteCompact

#End Region ' Dataset

#Region "DataTable"

		<Extension()>
		Public Function ClearCascade(Of T As DataTable)(table As T) As T ' clear childTables first
			Dim alreadyDone = New HashSet(Of DataTable)()
			Dim msg = New List(Of String)
			Dim recurse As Action(Of DataTable) =
			Sub(tb)
				msg.Add($"{tb.TableName}({tb.Rows.Count})")
				If tb.Rows.Count = 0 OrElse Not alreadyDone.Add(tb) Then Return 'prevent run in Cycles
				tb.ChildTables.ForEach(recurse)
				tb.Clear()
			End Sub
			If table.Rows.Count > 0 Then recurse(table) : Dbg("ClearCascade: " & ", ".Between(msg))
			Return table
		End Function

		''' <summary>Standard DataTable.Find unterstützt nur PrimaryKey-Columns</summary>
		<Extension()>
		Public Function FindX(table As DataTable, columns As DataColumn(), values As Object()) As DataRow
			If columns.SequenceEqual(table.PrimaryKey) Then Return table.Rows.Find(values)
			Return table.All.FirstOrDefault(Function(rw) columns.Select(Function(col) rw(col)).SequenceEqual(values))
		End Function

		<Extension()>
		Public Function ChangedRows(table As DataTable) As IEnumerable(Of DataRow)
			'DataRowState.Unchanged und DataRowState.Detached ausschließen
			Return From rw In table.Rows.Cast(Of DataRow)() Where rw.RowState >= DataRowState.Added
		End Function

		<Extension()>
		Public Function All(Of T As DataRow)(subj As TypedTableBase(Of T), Optional filter As Func(Of T, Boolean) = Nothing) As IEnumerable(Of T)
			All = DirectCast(subj, DataTable).All.Cast(Of T)()
			If filter.NotNull Then All = All.Where(filter)
		End Function
		<Extension()>
		Public Function All(subj As DataTable) As IEnumerable(Of DataRow)
			Return From rw In subj Where 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached))
		End Function

		<Extension()>
		Public Function ChildTables(subj As DataTable) As IEnumerable(Of DataTable)
			Return From rl In subj.ChildRelations.Cast(Of DataRelation)() Select rl.ChildTable
		End Function
		<Extension()>
		Public Function ParentTables(subj As DataTable) As IEnumerable(Of DataTable)
			Return subj.ParentRelations.Cast(Of DataRelation).Select(Function(rl) rl.ParentTable).Distinct
		End Function

		''' <summary>eases to implement standard-reactions on Row -init, -change, -delete</summary>
		<Extension()>
		Public Sub WireUp(Of T As DataRow)(table As TypedTableBase(Of T), Optional onNewRow As Action(Of T) = Nothing, Optional onChange As Action(Of T) = Nothing, Optional onDelete As Action(Of T) = Nothing)
			If onNewRow.NotNull() Then
				AddHandler table.TableNewRow, Sub(s, e) onNewRow(DirectCast(e.Row, T))
			End If
			If onChange.NotNull() Then
				AddHandler table.RowChanged,
						 Sub(s, e)
							 If 0 < (e.Action And (DataRowAction.Add Or DataRowAction.Change)) Then
								 onChange(DirectCast(e.Row, T))
							 End If
						 End Sub
			End If
			If onDelete.NotNull() Then
				AddHandler table.RowDeleting, Sub(s, e) onDelete(DirectCast(e.Row, T))
			End If
		End Sub

#End Region ' DataTable

#Region "DataRow"

		''' <summary>gibt es noch keine row mit denselben Schlüssel-Werten, so füge row ihrer DataTable hinzu - andernfalls
		''' setze row auf die gefundene, bereits vorhandene</summary>
		<Extension()>
		Public Sub EnsureExistUnique(Of T As DataRow)(ByRef row As T, ParamArray keyCols() As DataColumn)
			Dim tbl = DirectCast(row.Table, TypedTableBase(Of T))
			Dim rw = row
			If keyCols.Length = 0 Then keyCols = tbl.PrimaryKey
			Dim keyVals = (From col In keyCols Select col, val = rw(col)).ToArray
			rw = tbl.All.FirstOrDefault(Function(rw2) keyVals.All(Function(pv) rw2(pv.col).Equals(pv.val)))
			If rw.Null Then
				tbl.Rows.Add(row)
				If keyCols.Any(Function(clm) clm.AutoIncrement) Then row.AcceptChanges()
			Else
				row.Be(rw)
			End If
		End Sub

		<Extension()>
		Public Sub CopyTo(Of T As DataRow)(subj As T, other As T)
			Dim itms = subj.ItemArray
			Dim tb = other.Table
			For Each indx In tb.PrimaryKey.Select(Function(cl) cl.Ordinal)
				itms(indx) = other(indx)
			Next
			other.ItemArray = itms
		End Sub

#End Region' DataRow

	End Module
End Namespace