﻿#Region "FileHeader"
#If False Then
These BinarySearchers accept a selector-Delegate and a key-Value, to search Data with specific property matching to key.
Eg to search a FileInfo with specific Name in a List(Of FileInfo) you need no FileInfo with that Name, but simply can specify the searched Name.
   BinaryRangeSearcher bastardize the pure BinarySearch-Algorithm by never returning CompareResult 0 in Compare(x,y)-Method. In Effect the BinarySearch can't finish finding a match, but must search on to the final position **before** (or after) the first (last) Match. So I rapidly find the beginning of a range of matches.
#End If '-- Options, Imports
#End Region 'FileHeader

Namespace Global.System

	Public Enum SearcherType : FirstMatch = -1 : AnyMatch : LastMatch : End Enum

	Public MustInherit Class BinarySearcherBase(Of TElement, TProp) : Implements IComparer(Of TElement)
		Public Key As TProp

		Protected MustOverride Function CompareWithKey(x As TElement) As Integer

		Protected Overridable Function Compare(x As TElement, y As TElement) As Integer Implements IComparer(Of TElement).Compare
			Return CompareWithKey(x)
		End Function

		Public Overridable Function Search(lst As List(Of TElement), key As TProp, Optional index As Integer = 0, Optional count As Integer = -1) As Integer
			Me.Key = key
			If count < 0 Then count = lst.Count
			Return lst.BinarySearch(index, count, Nothing, Me)
		End Function
		Public Overridable Function Search(lst As TElement(), key As TProp, Optional index As Integer = 0, Optional count As Integer = -1) As Integer
			Me.Key = key
			If count < 0 Then count = lst.Length
			Return Array.BinarySearch(lst, index, count, Nothing, Me)
		End Function
	End Class

	Public MustInherit Class BinaryRangeSearcherBase(Of TElement, TProp) : Inherits BinarySearcherBase(Of TElement, TProp)

		Public SearcherType As SearcherType
		Protected _Match As Boolean

		Protected Overrides Function Compare(x As TElement, y As TElement) As Integer
			Dim retVal = CompareWithKey(x)
			If retVal = 0 Then
				_Match = True
				Return retVal - CInt(SearcherType)
			End If
			Return retVal
		End Function
		Public Overrides Function Search(lst As List(Of TElement), key As TProp, Optional index As Integer = 0, Optional count As Integer = -1) As Integer
			_Match = False
			Return CorrectResult(MyBase.Search(lst, key, index, count))
		End Function
		Public Overrides Function Search(lst As TElement(), key As TProp, Optional index As Integer = 0, Optional count As Integer = -1) As Integer
			_Match = False
			Return CorrectResult(MyBase.Search(lst, key, index, count))
		End Function
		Private Function CorrectResult(retVal As Integer) As Integer
			If _Match AndAlso SearcherType <> SearcherType.AnyMatch Then
				Return If(SearcherType = SearcherType.LastMatch, (Not retVal) - 1, Not retVal)
			End If
			Return retVal
		End Function
	End Class

	''' <summary>Führt binäre Suchen nach einem Key aus. Bei mehreren Treffern wird der erste/letzte gewählt. Der Key muss nicht IComparable sein, da die Comparison den Vergleich übernimmt.</summary>
	Public Class BinaryRangeComparisonSearcher(Of TElement, TProp) : Inherits BinaryRangeSearcherBase(Of TElement, TProp)
		Public ReadOnly Comparison As Func(Of TElement, TProp, Integer)
		Public Sub New(comparison As Func(Of TElement, TProp, Integer))
			Me.Comparison = comparison
		End Sub
		Protected Overrides Function CompareWithKey(x As TElement) As Integer
			Return Comparison(x, Key)
		End Function
	End Class
	''' <summary>Führt binäre Suchen nach einem (comparablen) Key aus. Bei mehreren Treffern wird der erste/letzte gewählt</summary>
	Public Class BinaryRangeSearcher(Of TElement, TProp As IComparable(Of TProp)) : Inherits BinaryRangeSearcherBase(Of TElement, TProp)
		Public ReadOnly Selector As Func(Of TElement, TProp)
		Public Sub New(selector As Func(Of TElement, TProp))
			Me.Selector = selector
		End Sub
		Protected Overrides Function CompareWithKey(x As TElement) As Integer
			Return Selector(x).CompareTo(Key)
		End Function
	End Class

	''' <summary>Führt binäre Suchen nach einem (comparablen) Key aus</summary>
	Public Class BinarySearcher(Of TElement, TProp As IComparable(Of TProp)) : Inherits BinarySearcherBase(Of TElement, TProp)
		Public ReadOnly Selector As Func(Of TElement, TProp)
		Public Sub New(selector As Func(Of TElement, TProp))
			Me.Selector = selector
		End Sub
		Protected Overrides Function CompareWithKey(x As TElement) As Integer
			Return Selector(x).CompareTo(Key)
		End Function
	End Class
	''' <summary>Führt binäre Suchen nach einem Key aus.  Der Key muss nicht IComparable sein, da die Comparison den Vergleich übernimmt.</summary>
	Public Class BinaryComparisonSearcher(Of TElement, TProp) : Inherits BinarySearcherBase(Of TElement, TProp)
		Public ReadOnly Comparison As Func(Of TElement, TProp, Integer)
		Public Sub New(comparison As Func(Of TElement, TProp, Integer))
			Me.Comparison = comparison
		End Sub
		Protected Overrides Function CompareWithKey(x As TElement) As Integer
			Return Comparison(x, Key)
		End Function
	End Class

	''' <summary>Führt binäre Suchen nach einem (comparablen) Key aus</summary>
	Public Class BinarySearcher
		''' <summary>Führt binäre Suchen nach einem (comparablen) Key aus. Bei mehreren möglichen Treffern wird je nach SearchType first/last/any gewählt.</summary>
		Public Shared Function Search(Of TElement, TProp As IComparable(Of TProp))(lst As List(Of TElement), key As TProp, selector As Func(Of TElement, TProp), Optional type As SearcherType = SearcherType.AnyMatch) As Integer
			If type = SearcherType.AnyMatch Then ' bei AnyMatch ist die Basisklasse performanter
				Dim searcher = New BinarySearcher(Of TElement, TProp)(selector) With {.Key = key}
				Return searcher.Search(lst, key)
			End If
			Dim brs = New BinaryRangeSearcher(Of TElement, TProp)(selector) With {.SearcherType = type}
			Return brs.Search(lst, key)
		End Function

		''' <summary>Führt binäre Suchen nach einem (comparablen) Key aus. Bei mehreren möglichen Treffern wird je nach SearchType first/last/any gewählt.</summary>
		Public Shared Function Search(Of TElement, TProp As IComparable(Of TProp))(lst As TElement(), key As TProp, selector As Func(Of TElement, TProp), Optional type As SearcherType = SearcherType.AnyMatch) As Integer
			If type = SearcherType.AnyMatch Then ' bei AnyMatch ist die Basisklasse performanter
				Dim searcher = New BinarySearcher(Of TElement, TProp)(selector) With {.Key = key}
				Return searcher.Search(lst, key)
			End If
			Dim brs = New BinaryRangeSearcher(Of TElement, TProp)(selector) With {.SearcherType = type}
			Return brs.Search(lst, key)
		End Function

		''' <summary>Führt binäre Suchen nach einem Key aus. Bei mehreren möglichen Treffern wird je nach SearchType first/last/any gewählt.</summary>
		Public Shared Function Search(Of TElement, TProp)(lst As TElement(), key As TProp, comparison As Func(Of TElement, TProp, Integer), Optional type As SearcherType = SearcherType.AnyMatch) As Integer
			If type = SearcherType.AnyMatch Then ' bei AnyMatch ist die Basisklasse performanter
				Dim searcher = New BinaryComparisonSearcher(Of TElement, TProp)(comparison) With {.Key = key}
				Return searcher.Search(lst, key)
			End If
			Dim brs = New BinaryRangeComparisonSearcher(Of TElement, TProp)(comparison) With {.SearcherType = type}
			Return brs.Search(lst, key)
		End Function

		''' <summary>Führt binäre Suchen nach einem Key aus. Bei mehreren möglichen Treffern wird je nach SearchType first/last/any gewählt.</summary>
		Public Shared Function Search(Of TElement, TProp)(lst As List(Of TElement), key As TProp, comparison As Func(Of TElement, TProp, Integer), Optional type As SearcherType = SearcherType.AnyMatch) As Integer
			If type = SearcherType.AnyMatch Then ' bei AnyMatch ist die Basisklasse performanter
				Dim searcher = New BinaryComparisonSearcher(Of TElement, TProp)(comparison) With {.Key = key}
				Return searcher.Search(lst, key)
			End If
			Dim brs = New BinaryRangeComparisonSearcher(Of TElement, TProp)(comparison) With {.SearcherType = type}
			Return brs.Search(lst, key)
		End Function

	End Class

End Namespace
